#!/usr/bin/python
# -*- coding: utf8 -*-

##
# @brief hiairv2 api for python
#
#    1.依https://sms.hinet.net 軟體開發規格撰寫
#    2.此原始檔案為UTF8編碼，所以取multibyte時為UTF8
#
#    2016/5/6 初版，支援長簡訊
#    2021/01/18 修正版本適用python3.9

import math
import socket
import struct
from utils.encoding_lib import is_english

# 連線資訊
PORT = 8000

# 資料長度
# 發送結構長度
SEND_MSG_LENGTH = 266
SEND_SET_LEN = 100
# 接收結構長度
RET_MSG_LENGTH = 244
RET_SET_LEN = 80
# 共用
CONTENT_LEN = 160

# 長簡訊分割長度
# 英數字153
MAX_LONG_ENGLEN = 153
# 非英數字67
MAX_LONG_CHILEN = 67
MAX_LONG_LEN = 10

# 逾時設定(勿更改)
MAX_TIMEOUT_SECONDS = 60

# 錯誤代碼
SOCKET_CONN_ERROR = -1
SOCKET_SEND_RECV_ERROR = -2
EXCEED_MAX_LEN_ERROR = -9  # 超過長簡訊最大分則數(10則)

# 編碼設定
# 1 : big5
# 2 : binary (目前不使用)
# 3 : unicode(UCS-2 / UTF-16，採Big Endian，即UCS-2BE、UTF-16BE)
# 4 : unicode(UTF-8) 若用UTF-8送的話，70個中文字只會顯示46個，因為UTF-8 非英數字占用3個byte,UCS-2占用2個byte，建議使用UCS-2。
CODING_BIG5 = 1
CODING_UCS2 = 3
CODING_STR_BIG5 = "BIG5"
CODING_STR_UCS2 = "UTF-16BE"

# 訊息型態
SEND_LOGIN = 0
SEND_TEXT = 1
SEND_FOREIGN = 15
SEND_LONG = 11
SEND_FOREIGN_LONG = 18

# 傳送型式
SEND_NOW = 1
SEND_NOW_EXPIRED = 2
SEND_RESERVE = 3
SEND_RESERVE_EXPIRED = 4


class Hiairv2:

    ##
    # @brief 初始化
    #
    # @param sock
    #
    # @return
    def __init__(self, sock=None):

        if sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.settimeout(MAX_TIMEOUT_SECONDS)
        else:
            self.sock = sock

    ##
    # @brief 建立連線並驗證帳號密碼
    #
    # @param server 伺服器IP(202.39.54.130)
    # @param username 使用者帳號
    # @param password 使用者密碼
    #
    # @return (ret_code:回應代碼, ret_content:回應文字)
    def StartCon(self, server, username, password):

        try:
            # socket連線
            self.sock.connect((server, PORT))

            # 組msg_set，填帳號密碼
            msg_set = username + "\0" + password + "\0"
            msg_set = msg_set.encode(CODING_STR_BIG5)
            msg_set_len = len(msg_set)
            msg_set = msg_set.ljust(SEND_SET_LEN, "\0".encode(CODING_STR_BIG5))  # 要補滿，超過要截斷補\0

            # 組msg_content
            msg_content = ""
            msg_content = msg_content.encode(CODING_STR_BIG5)
            msg_content_len = len(msg_content)
            msg_content = msg_content.ljust(CONTENT_LEN, "\0".encode(CODING_STR_BIG5))

            # 傳送
            send_data = self.GenSendData(SEND_LOGIN, CODING_BIG5, 0, 0, msg_set_len, msg_content_len, msg_set,
                                         msg_content)
            sent = self.sock.send(send_data)

            # 接收
            recv_data = self.sock.recv(RET_MSG_LENGTH)

            # 回傳
            return self.GetRecvData(recv_data)
        except Exception as e:
            return SOCKET_CONN_ERROR, "StartCon() socket create error: " + str(e)

    ##
    # @brief 發送短簡訊
    #
    # @param mobileNum 手機號碼
    # @param message 簡訊內容
    # @param send_type 傳送型態
    # @param limit_time 重送截止時間
    # @param order_time 預約時間
    #
    # @return (ret_code:回應代碼, ret_content:回應文字)
    def SendTextMessage(self, mobileNum, message, send_type=1, limit_time=1440, order_time=""):

        # 國內外簡訊判斷
        msg_type = None
        if ((mobileNum[0:1] == "+") and (mobileNum[0:4] != "+886")):
            msg_type = SEND_FOREIGN  # 國外門號
        else:
            msg_type = SEND_TEXT  # 國內門號 #09 +8869 +88609
            # 手機格式調整
            mobileNum = (mobileNum.replace("+88609", "+8869", 1)).replace("+8869", "09", 1)

        # 檢查重送截止時間
        if ((limit_time < 0) or (limit_time > 1440)):
            limit_time = 1440

        # 傳送型態與重送截止時間轉為固定長度字串
        stype = "{:02}".format(send_type)
        limit = "{:04}".format(limit_time)

        # 組msg_set
        msg_set = None
        if send_type == SEND_NOW:  # 立即
            msg_set = mobileNum + '\0' + stype + '\0'
        elif send_type == SEND_NOW_EXPIRED:  # 立即 + 重送截止
            msg_set = mobileNum + '\0' + stype + '\0' + limit + '\0'
        elif send_type == SEND_RESERVE:  # 預約
            msg_set = mobileNum + '\0' + stype + '\0' + order_time + '\0'
        elif send_type == SEND_RESERVE_EXPIRED:  # 預約+重送截止
            msg_set = mobileNum + '\0' + stype + '\0' + limit + '\0' + order_time + '\0'
        else:  # 有問題的，就以立即發送傳送
            msg_set = mobileNum + '\0' + "{:02}".format(SEND_NOW) + '\0'

        # 組msg_set
        msg_set = msg_set.encode(CODING_STR_BIG5)
        msg_set_len = len(msg_set)
        msg_set = msg_set.ljust(SEND_SET_LEN, "\0".encode(CODING_STR_BIG5))

        # 編碼
        msg_coding = None

        # 如果utf8的byte長度與原始字串長度相同，為純英數字
        if is_english(message):  # 純英數字每則159，編碼用big5
            message = message[:159]  # 長度要<=159
            msg_coding = CODING_BIG5
            coding_str = CODING_STR_BIG5
        else:  # 非英數字每則70，編碼用UCS-2BE
            message = message[:70]  # 長度要<=70
            msg_coding = CODING_UCS2
            coding_str = CODING_STR_UCS2

        # 組msg_content，原始檔案為 utf8編碼，轉換簡訊內容為Big5 或 UCS2(即UTF-16BE)編碼
        message = message.encode(coding_str)
        msg_content_len = len(message)
        msg_content = message + "\0".encode(CODING_STR_BIG5) * (CONTENT_LEN - msg_content_len)

        recv_data = None
        try:
            # 送出
            send_data = self.GenSendData(msg_type, msg_coding, 0, 0, msg_set_len, msg_content_len, msg_set, msg_content)
            sent = self.sock.send(send_data)
            # 接收
            recv_data = self.sock.recv(RET_MSG_LENGTH)
        except Exception as e:
            return (SOCKET_SEND_RECV_ERROR, "SendTextMessage() socket error: " + str(e))

        # 回傳
        return self.GetRecvData(recv_data)

    ##
    # @brief 發送長簡訊
    #
    # @param mobileNum 手機號碼
    # @param message 簡訊內容
    # @param send_type 傳送型態
    # @param limit_time 重送截止時間
    # @param order_time 預約發送時間
    #
    # @return (ret_code:回應代碼, ret_content:回應文字)
    def SendLongMessage(self, mobileNum, message, send_type=1, limit_time=1440, order_time=""):

        # 分則訊息長度
        msg_size_limit = None
        # 編碼
        msg_coding = None

        # 如果utf8的byte長度與原始字串長度相同，為純英數字
        if is_english(message):
            msg_size_limit = MAX_LONG_ENGLEN  # 純英數字每則153，編碼用big5
            msg_coding = CODING_BIG5
            coding_str = CODING_STR_BIG5
        else:
            msg_size_limit = MAX_LONG_CHILEN  # 非英數字每則67，編碼用UCS-2BE
            msg_coding = CODING_UCS2
            coding_str = CODING_STR_UCS2

        # 計算要多少通分則簡訊
        numOfMsg = math.ceil(len(message) / float(msg_size_limit))

        # 超過最大分則(10則)
        if (numOfMsg > MAX_LONG_LEN):
            return (EXCEED_MAX_LEN_ERROR, "SendLongMessage() exceed max sms number limit(10)")

        # 檢查重送截止時間
        if (limit_time < 1 and limit_time > 1440):
            limit_time = 1440

        # 傳送型態與重送截止時間轉為固定長度字串
        stype = "{:02}".format(send_type)
        limit = "{:04}".format(limit_time)

        for cnt in range(0, int(numOfMsg + 1)):

            if cnt == 0:  # 第一通，傳送長簡訊資訊

                # 國內外簡訊判斷
                msg_type = None
                if ((mobileNum[0:1] == "+") and (mobileNum[0:4] != "+886")):
                    msg_type = SEND_FOREIGN_LONG
                else:
                    msg_type = SEND_LONG
                    # 手機格式調整
                    mobileNum = (mobileNum.replace("+88609", "+8869", 1)).replace("+8869", "09", 1)

                msg_set = None
                if send_type == SEND_NOW:  # 立即
                    msg_set = mobileNum + '\0' + stype + '\0'
                elif send_type == SEND_NOW_EXPIRED:  # 立即+重送截止
                    msg_set = mobileNum + '\0' + stype + '\0' + limit + '\0'
                elif send_type == SEND_RESERVE:  # 預約
                    msg_set = mobileNum + '\0' + stype + '\0' + order_time + '\0'
                elif send_type == SEND_RESERVE_EXPIRED:  # 預約+重送截止
                    msg_set = mobileNum + '\0' + stype + '\0' + limit + '\0' + order_time + '\0'
                else:  # 有問題的，改以立即傳送發出
                    msg_set = mobileNum + '\0' + "{:02}".format(SEND_NOW) + '\0'

                # 組msg_set
                msg_set = msg_set.encode(CODING_STR_BIG5)
                msg_set_len = len(msg_set)
                msg_set = msg_set.ljust(SEND_SET_LEN, "\0".encode(CODING_STR_BIG5))

                msg_content_len = int(numOfMsg)  # 填總通數
                msg_content = "\0".encode(CODING_STR_BIG5) * CONTENT_LEN

                recv_data = None
                try:
                    # 傳送
                    send_data = self.GenSendData(msg_type, msg_coding, 0, 0, msg_set_len, msg_content_len, msg_set,
                                                 msg_content)
                    sent = self.sock.send(send_data)

                    # 接收
                    recv_data = self.sock.recv(RET_MSG_LENGTH)
                except Exception as e:
                    return (SOCKET_SEND_RECV_ERROR, "SendLongMessage0() socket error:" + str(e))

                (ret_code, ret_content) = self.GetRecvData(recv_data)

                # 如果分則有錯誤，退出
                if (ret_code != 0):
                    return (ret_code, ret_content)

            else:  # 第一通以後，傳送長簡訊內容

                msg_type = cnt  # 第幾則
                msg_set = "\0".encode(CODING_STR_BIG5) * SEND_SET_LEN
                msg_set_len = 1  # length for '\0'

                # 取分則子字串
                part_message = message[(cnt - 1) * msg_size_limit:cnt * msg_size_limit]

                # 轉換編碼
                buf_message = None
                buf_message = part_message.encode(coding_str)  # 純英數字以BIG5/ASCII送出 # 非純英數字以UCS-2BE/UTF-16BE送出
                msg_content_len = len(buf_message)
                msg_content = buf_message + "\0".encode(CODING_STR_BIG5) * (CONTENT_LEN - msg_content_len)

                recv_data = None
                try:
                    # 傳送
                    send_data = self.GenSendData(msg_type, msg_coding, 0, 0, msg_set_len, msg_content_len, msg_set,
                                                 msg_content)
                    sent = self.sock.send(send_data)

                    # 接收
                    recv_data = self.sock.recv(RET_MSG_LENGTH)
                except Exception as e:
                    return (SOCKET_SEND_RECV_ERROR, "SendLongMessage+() socket error:" + str(e))

                (ret_code, ret_content) = self.GetRecvData(recv_data)

                # 若傳送有誤，或是最後一筆 => 回傳結果
                if ((ret_code != 0) or (cnt == numOfMsg)):
                    return (ret_code, ret_content)

    ##
    # @brief 查詢短簡訊發送狀態
    #
    # @param messageID MessageID
    #
    # @return  (ret_code:回應代碼, ret_content:回應文字)
    def QueryMessage(self, message_id):
        return self.Query(2, message_id)

    ##
    # @brief 查詢長簡訊發送狀態
    #
    # @param messageID MessageID
    #
    # @return  (ret_code:回應代碼, ret_content:回應文字)
    def QueryLongMessage(self, message_id):
        return self.Query(12, message_id)

    ##
    # @brief 查詢簡訊狀態
    #        此為實作，直接用QueryMessage()或QueryLongMessage()
    # @param msg_type 訊息型態
    # @param messageID MessageID
    #
    # @return  (ret_code:回應代碼, ret_content:回應文字)
    def Query(self, msg_type, message_id):

        # 組msg_set，填messageID
        msg_set = (message_id + "\0").encode(CODING_STR_BIG5)
        msg_set_len = len(msg_set)
        msg_set = msg_set.ljust(SEND_SET_LEN, "\0".encode(CODING_STR_BIG5))

        # 組msg_content
        msg_content = "".encode(CODING_STR_BIG5)
        msg_content_len = len(msg_content)
        msg_content = msg_content.ljust(CONTENT_LEN, "\0".encode(CODING_STR_BIG5))

        recv_data = None
        try:
            # 傳送
            send_data = self.GenSendData(msg_type, CODING_BIG5, 0, 0, msg_set_len, msg_content_len, msg_set,
                                         msg_content)
            sent = self.sock.send(send_data)

            # 接收
            recv_data = self.sock.recv(RET_MSG_LENGTH)

        except Exception as e:
            return (SOCKET_SEND_RECV_ERROR, "Query() socket error:" + str(e))

        # 回傳
        return self.GetRecvData(recv_data)

    ##
    # @brief 取消 預約短簡訊
    #
    # @param message_id MessageID
    #
    # @return  (ret_code:回應代碼, ret_content:回應文字)
    def CancelMessage(self, message_id):
        return self.Cancel(16, message_id)

    ##
    # @brief 取消預約長簡訊
    #
    # @param message_id MessageID
    #
    # @return  (ret_code:回應代碼, ret_content:回應文字)
    def CancelLongMessage(self, message_id):
        return self.Cancel(17, message_id)

    def Cancel(self, msg_type, message_id):
        # 組msg_set，填messageID
        msg_set = (message_id + "\0").encode(CODING_STR_BIG5)
        msg_set_len = len(msg_set)
        msg_set = msg_set.ljust(SEND_SET_LEN, "\0".encode(CODING_STR_BIG5))

        # 組msg_content
        msg_content = "".encode(CODING_STR_BIG5)
        msg_content_len = len(msg_content)
        msg_content = msg_content.ljust(CONTENT_LEN, "\0".encode(CODING_STR_BIG5))

        recv_data = None
        try:
            # 傳送
            send_data = self.GenSendData(msg_type, CODING_BIG5, 0, 0, msg_set_len, msg_content_len, msg_set,
                                         msg_content)
            sent = self.sock.send(send_data)

            # 接收
            recv_data = self.sock.recv(RET_MSG_LENGTH)

        except Exception as e:
            return (SOCKET_SEND_RECV_ERROR, "Cancel() socket error:" + str(e))

        # 回傳
        return self.GetRecvData(recv_data)

    ##
    # @brief 組合要送出的訊息結構
    #
    # @param msg_type
    # @param msg_coding
    # @param msg_priority
    # @param msg_country_code
    # @param msg_set_len
    # @param msg_content_len
    # @param msg_set
    # @param msg_content
    #
    # @return
    def GenSendData(self, msg_type, msg_coding, msg_priority, msg_country_code, msg_set_len, msg_content_len, msg_set,
                    msg_content):
        return struct.pack("B", msg_type) + struct.pack("B", msg_coding) + struct.pack("B", msg_priority) + \
               struct.pack("B", msg_country_code) + struct.pack("B", msg_set_len) + \
               struct.pack("B", msg_content_len) + msg_set + msg_content

    ##
    # @brief 取出回傳的訊息結構
    #
    # @param recv_data
    #
    # @return
    def GetRecvData(self, recv_data):
        ret_code = struct.unpack("B", recv_data[0:1])[0]
        # ret_coding = struct.unpack("B", recv_data[1:2])[0]
        # ret_set_len = struct.unpack("B", recv_data[2:3])[0]
        # ret_length = struct.unpack("B", recv_data[3:4])[0]
        # ret_set = recv_data[4: 4 + RET_SET_LEN + 1]
        ret_content = recv_data[84:RET_MSG_LENGTH + 1]
        return ret_code, (ret_content.decode(CODING_STR_BIG5)).strip()

    ##
    # @brief 閉關socket連線
    #
    # @return
    def EndCon(self):
        self.sock.close()
